#include <SoftwareSerial.h>             //Communication with GPS module
#include <stdlib.h>                     //String to integer/float conversion
#include <TM1637Display.h>              //Used for 7-segment displays
#include <EEPROM.h>                     //Saving GPS coodinates to non-volatile memory so we can cut the power whislt retaining our last know position

#define ESCButton 4
#define SELECTButton 3

#define LED_Strip_R 10                  //Red cathode of RGB LED strips
#define LED_Strip_G 13                   //Green cathode of RGB LED strips

#define RED_LED 6                       //LED is red when no GPS data avaliable 
#define GREEN_LED 5                     //LED is green when GPS data avaliable 
#define BuzzerPin 7

#define GPS_RXPin 12                   
#define GPS_TXPin 11
#define GPSBaudRate 9600
#define GPSPowerPin 2                   //A pin connected to a transitor which gives the GPS power when it's needed

#define SegClock 9
#define SegData 8

#define BUFFSIZE 100                    //Size of an array used to store GPS transmissions (The first 100 characters usually contain useful data)
char buffer[BUFFSIZE];
int bufIndex;
int result;                             //Stores lenth of transmission
                                        //Arrays for storing location data
char latDeg[3];
char longDeg[4];
char latTime[9];
char longTime[9];
                                        //Variables for when arrays are converted to numbers
int latDeg_int;
int longDeg_int;
float latTime_float;
float longTime_float;
                                        //Variables for the decimal longitude and latitude
int DataReceived=0;                     //This lets us know later if we have anything useful we can save to EEPROM if requested                                
float decimal_lat;
float decimal_long;
float decimal_latPAST;
float decimal_longPAST;

int runSubprogram;                      //Variable used to control weather or not a function should loop
int programNumber;                      //Variable to decide which fucntion to use (i.e. program 0, program 1, program 2)

TM1637Display display1 = TM1637Display(SegClock,SegData); //Define the clock and data pins for our display
SoftwareSerial gpsSerial(GPS_RXPin, GPS_TXPin); //Begins transmission on declated pins

const uint8_t SEG_END[]={               //Defining an arrangement of segments to display "End_" on the 7-segment display    
  SEG_A | SEG_F | SEG_G | SEG_E | SEG_D,
  SEG_E | SEG_G | SEG_C,
  SEG_B | SEG_C | SEG_D | SEG_E | SEG_G,
  SEG_D
};

const uint8_t SEG_SAFE[]={               //Defining an arrangement of segments to display "SAFE" on the 7-segment display    
  SEG_A | SEG_F | SEG_G | SEG_C | SEG_D,
  SEG_E | SEG_F | SEG_A | SEG_B | SEG_C | SEG_G,
  SEG_E | SEG_F | SEG_G | SEG_A,
  SEG_A | SEG_F | SEG_G | SEG_E | SEG_D
};

const uint8_t SEG_ERRO[]={               //Defining an arrangement of segments to display "ERRO" on the 7-segment display    
  SEG_A | SEG_F | SEG_G | SEG_E | SEG_D,
  SEG_A | SEG_F | SEG_E,
  SEG_A | SEG_F | SEG_E,
  SEG_E | SEG_F | SEG_A | SEG_B | SEG_C | SEG_D
};

const uint8_t SEG_DATA[]={               //Defining an arrangement of segments to display "dAtA" on the 7-segment display    
  SEG_B | SEG_C | SEG_D | SEG_E | SEG_G,
  SEG_E | SEG_F | SEG_A | SEG_B | SEG_C | SEG_G,
  SEG_E | SEG_G | SEG_D | SEG_F,
  SEG_E | SEG_F | SEG_A | SEG_B | SEG_C | SEG_G
};
//----------------------------------------auditoryLocator----------------------------------------------------------------------------------

//FUN FACT: This function has two sound types, the first acts a white noise which makes nearby location easy as you brain can sync the sound waves entering your ears by turing 
//your head, this cool video explains this phenomenom better https://www.youtube.com/watch?v=CqB95rj_txI , the second sound type is a simple sine wave which can be
//heard from further away as the wave is very distinctive and high frequency.

void auditoryLocator() {                  //Beeps and flashes, acting as an auditory locator
  int soundType=1;
  while(runSubprogram==1){
    digitalWrite(LED_Strip_R,HIGH);
    if(soundType==1){
      digitalWrite(GREEN_LED,HIGH);
      digitalWrite(RED_LED,HIGH );
      for(int i=0;i<2500;i++){            //Produce white noise for approximatly 1 second
      digitalWrite(BuzzerPin, HIGH);
      delayMicroseconds(random(1, 100));
      digitalWrite(BuzzerPin, LOW);
      delayMicroseconds(random(1, 100));
      soundType=0;
      }
      digitalWrite(GREEN_LED,LOW);
      digitalWrite(RED_LED,LOW);
    }else{
      digitalWrite(BuzzerPin,HIGH);
      digitalWrite(GREEN_LED,HIGH);
       digitalWrite(RED_LED,HIGH);
      delay(500);
      digitalWrite(BuzzerPin,LOW);
      digitalWrite(GREEN_LED,LOW);
      digitalWrite(RED_LED,LOW);
      soundType=1;
    }
    if(digitalRead(ESCButton)==LOW){
      initiateClose();
    }
    digitalWrite(LED_Strip_R, LOW);
    delay(1000);   
  }
}
//----------------------------------------GPS---------------------------------------------------------------------------------------------

void GPS(){
  digitalWrite(GPSPowerPin,HIGH);            //Give power to the transistor managing the GPS module to turn it on
  while(runSubprogram==1){
    while (gpsSerial.available() > 0){
      char c = gpsSerial.read();
        
     if (c == '\n') {                        //If the GPS finishes sending a message
        buffer[bufIndex]='\0';               //Terminate the string with a null character
        Serial.println(buffer);
        result = checkString(buffer);        //Find the length of the transmission
        bufIndex=0;
        if(result>38){                       //If long enough, we can assume the transmission has GPS data
          digitalWrite(GREEN_LED,HIGH);
          digitalWrite(RED_LED,LOW);
          DataReceived=1;                    //This lets us know later if we have anything useful we can save to EEPROM if requested
          Decimal_LongLat(buffer);           //Converts transmission into decimal degrees
        }else
        if(result<38 && result>0){          //If we recieve a message starting "$GPGLL but its too short, we try to connect
          digitalWrite(RED_LED,HIGH);
          digitalWrite(GREEN_LED,LOW);
        }       
      }else{
        if (bufIndex<(BUFFSIZE-1)) buffer[bufIndex++]=c; //If theres room, store the incoming character to the array but always leave room for a null character
      }  
    }
    if(digitalRead(ESCButton)==LOW){      
      initiateClose(); 
    }
  }
  digitalWrite(GPSPowerPin,LOW);        //Shut down the GPS
  digitalWrite(GREEN_LED,LOW);
  digitalWrite(RED_LED,LOW);
}

int checkString(char * b) {             //Finds the length of strings starting "$GPGLL" (loction data)
  if (0==(strncmp("$GPGLL",b,6))){
    return strlen(b);
  }else{
    return 1;
  }
}

void Decimal_LongLat(char*S){            //Converts GPS transmission to decimal degrees
  for(int i=0;i<4;i++){
    if(i==3){
      longDeg[i]='\0';
      longDeg_int=atoi(longDeg);         //Converts Ascii aray to integer
    }else{
    longDeg[i]=S[i+20];                  //+20 goes to character index 20 in the transmission before recording data
    }                                    //which gives us the data we want (here its the Longitudinal degrees)
  }
  for(int i=0;i<3;i++){
    if(i==2){
      latDeg[i]='\0';
      latDeg_int=atoi(latDeg);
    }else{
    latDeg[i]=S[i+7];
    }
  }
  for(int i=0;i<9;i++){
    if(i==8){
      longTime[i]='\0';
      longTime_float=atof(longTime);      //Converts Ascii array to float value
    }else{
    longTime[i]=S[i+23];
    }
  }
  for(int i=0;i<9;i++){
    if(i==8){
      latTime[i]='\0';
      latTime_float=atof(latTime);
    }else{
    latTime[i]=S[i+9];
    }
  }
  decimal_lat=latDeg_int+(latTime_float/60); //Calculates decimal latitude 
  decimal_long=longDeg_int+(longTime_float/60); //Calculate decimal longitude
  if(S[18]=='S'){                         //If in Southern hemisphere, deimal Latitude is a negative
    decimal_lat=-decimal_lat;
  }
  if(S[32]=='W'){                         //If in West side, deimal longitude is a negative
    decimal_long=-decimal_long;
  }
  Serial.println("------------------------------");
  Serial.print(decimal_lat,6);            //Prints to 6 decimal places
  Serial.print(",");
  Serial.println(decimal_long,6);
  Serial.println("------------------------------");
  bearing();
}

//----------------------------------------SubProgramClose----------------------------------------------------------------------------------

void initiateClose(){                     //Sets "runSubprogram" to 0 so all subrograms are closed
  display1.clear();
  display1.setSegments(SEG_END);
  Serial.println("close initiated");
  digitalWrite(BuzzerPin,HIGH);
  delay(750);
  digitalWrite(BuzzerPin,LOW);
  runSubprogram=0; 
}

//----------------------------------------simonSays----------------------------------------------------------------------------------------

short sequence[100]={};
void simonSays(){
  randomSeed(millis());
  while(runSubprogram==1){
    for(int i=0;i<100;i++){
      sequence[i]=random(0,2);
    }
    for(int j=0;j<100;j++){
      for(int i=0;i<j;i++){
        display1.showNumberDec(i+1);
        if(sequence[i]==0){
          digitalWrite(LED_Strip_R,HIGH);
        }else{
          digitalWrite(LED_Strip_G,HIGH);
        }      
        for(int i=0;i<4000;i++){
          digitalWrite(BuzzerPin, HIGH);
          delayMicroseconds(random(1, 100));
          digitalWrite(BuzzerPin, LOW);
          delayMicroseconds(random(1, 100));
        }
        digitalWrite(LED_Strip_R,LOW);
        digitalWrite(LED_Strip_G,LOW);
        delay(200);
      }
    }
  }
}
//----------------------------------------Bearing--------------------------------------------------------------------------------------------

void bearing() {                    

   union {
    float floatData;
    char bytes[4];
  } unionEEPROM;
  
  //Reading from EEPROM
  for (int i = 0; i < 4; ++i) unionEEPROM.bytes[i] = EEPROM.read(i);
  decimal_latPAST = unionEEPROM.floatData;            //Read the first 4 bytes of EEPROM and set that to out past lat coordinate
  Serial.println(decimal_latPAST, 6);

  for (int i = 4; i < 8; ++i) unionEEPROM.bytes[i-4] = EEPROM.read(i);
  decimal_longPAST = unionEEPROM.floatData;           //Read the next 4 bytes of EEPROM and set that to out past long coordinate
  Serial.println(decimal_longPAST, 6);

  float teta1 = radians(decimal_lat);
  float teta2 = radians(decimal_latPAST);
  float delta1 = radians(decimal_latPAST - decimal_lat);
  float delta2 = radians(decimal_longPAST - decimal_long);

  //HeadingFormula

  float y = sin(delta2) * cos(teta2);
  float x = cos(teta1)*sin(teta2) - sin(teta1)*cos(teta2)*cos(delta2);
  float bearing = atan2(y,x);
  bearing = degrees(bearing);// radians to degrees
  bearing = fmod((bearing + 360), 360);
  Serial.print("Heading GPS: ");
  Serial.println(bearing);
  int abc = bearing;
  display1.showNumberDec(abc);
  
  if(digitalRead(ESCButton)==LOW){
    initiateClose();
  }
}


//----------------------------------------EEPROMSave-----------------------------------------------------------------------------------------

void EEPROMSave() {
  union {
    float floatData;
    char bytes[4];
  } unionEEPROM;

  if(DataReceived==0){                    //if we have no new data, display an error message
    display1.setSegments(SEG_ERRO);
    digitalWrite(BuzzerPin,HIGH);
    digitalWrite(RED_LED,HIGH);
    delay(750);
    digitalWrite(BuzzerPin,LOW);
    digitalWrite(RED_LED,LOW);
    delay(500);
    display1.setSegments(SEG_DATA);
    delay(1000);
  }else{
    int Run=1;                            //Otherwise start a countdown for EEPROM saving. Count down allows you to cancel so you dont wipe your last know location.
    for(int i=0;(i<6)&&Run==1;i++){
      display1.showNumberDec(5-i);
      digitalWrite(BuzzerPin,HIGH);
      digitalWrite(RED_LED,HIGH);
      if(digitalRead(ESCButton)==LOW){
        Run=0;
        display1.setSegments(SEG_END);}
      delay(500);
      digitalWrite(BuzzerPin,LOW);
      digitalWrite(RED_LED,LOW);
      delay(500);
      }
         
    if(Run==1){
      Serial.println(decimal_lat,6);
      unionEEPROM.floatData = decimal_lat;
      for (int i = 0; i < 4; ++i) EEPROM.write(i, unionEEPROM.bytes[i]);    //Save the 4 bytes to the first four adresses
    
      Serial.println(decimal_long,6);
      unionEEPROM.floatData = decimal_long;
      for (int i = 4; i < 8; ++i) EEPROM.write(i, unionEEPROM.bytes[i-4]);  //Save the next 4 bytes to the next four adresses
    
      display1.setSegments(SEG_SAFE);     //Display a confimration message
      DataReceived=0;
      Run=0;
      for(int i=0;i<5;i++){               // Beep to confirm data saved
        digitalWrite(BuzzerPin,HIGH);
        digitalWrite(GREEN_LED,HIGH);
        delay(100);
        digitalWrite(BuzzerPin,LOW);
        digitalWrite(GREEN_LED,LOW);
        delay(100);
      }
    }   
    delay(1000);
  }
}

//----------------------------------------Setup&Main---------------------------------------------------------------------------------------

void setup() {
  display1.setBrightness(5);                //Set the 7-segment display brightness
  pinMode(GPSPowerPin,OUTPUT);
  pinMode(RED_LED,OUTPUT);
  pinMode(GREEN_LED,OUTPUT);
  pinMode(ESCButton, INPUT_PULLUP);
  pinMode(SELECTButton, INPUT_PULLUP);
  Serial.begin(9600);
  gpsSerial.begin(GPSBaudRate);
}

//----------------------------------------FunctionDecleration------------------------------------------------------------------------------
void (*functionArray[])() = {GPS, auditoryLocator,EEPROMSave}; //An array containting the names of subprograms which we run through in the menu, you can replace these for your own functions
int NumPrograms = sizeof(functionArray) / sizeof(functionArray[0]);

void loop() {
  display1.showNumberDec(programNumber+1);  //+1 so that the list starts at 1, just for a (personally) nicer menu
  if(digitalRead(ESCButton)==LOW){
    if(programNumber<NumPrograms-1){
      programNumber++;
    }else
      programNumber=0;  
      while(digitalRead(ESCButton)==LOW);   //Waiting for user to let go of button
      
  }else 
    if(digitalRead(SELECTButton)==LOW){
      runSubprogram=1;
      (*functionArray[programNumber])();
      while(digitalRead(SELECTButton)==LOW);  //Waiting for user to let go of button
   } 
}
